package dev.pfaff.altargui.widgets.state;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.inspector.InspectorMeta;
import dev.pfaff.altargui.selector.BuiltinSelector;
import dev.pfaff.altargui.selector.Children;
import dev.pfaff.altargui.selector.NthElement;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;
import java.util.function.Consumer;

public abstract class SingleChildWidgetInstance<W extends Widget> extends WidgetInstance<W> {
	@InspectorMeta(tooltip = "The currently mounted child element")
	private WidgetInstance<?> childElement;

	protected SingleChildWidgetInstance(WidgetInstantiationArgs args) {
		super(args);
	}

	protected final void setChild(StateContext stateContext, @NotNull Widget child) {
		setChild(stateContext, child, true);
	}

	protected Key.CallSite childKey() {
		return () -> {};
	}

	protected final void setChild(StateContext stateContext, @NotNull Widget child, boolean dirtyStencil) {
		var element = stateContext.getState(this, this.childKey(), Objects.requireNonNull(child));
		if (this.childElement != element) {
			this.childElement = element;
			if (dirtyStencil) {
				markNeedsStencilLayout();
			}
			markNeedsTraceLayout();
			markNeedsScissorLayout();
		}
	}

	protected final WidgetInstance<?> getChildElement() {
		var childElement = this.childElement;
		assert childElement != null : STR."child of \{toStringShort()} is not initialized:\n" + this;
		return childElement;
	}

	@Override
	protected void computeStencilLayout(StateContext context) {
		WidgetInstance<?> element = getChildElement();
		Objects.requireNonNull((WidgetInstance<?>) this, "Parent must be non-null");
		Objects.requireNonNull(element, "Element must be non-null");

		element.attachTo(this);
		assert element.parent() != null;

		element.copyAvailableSizeFrom(this);
		element.stencilLayout(context);
		copyComputedSizeFrom(element);
	}

	@Override
	protected void computeTraceLayout() {
		var e = getChildElement();
		e.copyPositionFrom(this);
		e.traceLayout();
	}

	@Override
	protected void computeScissorLayout() {
		this.childElement.copyClippedRectFrom(this);
		this.childElement.scissorLayout();
	}

	@Override
	protected void computeMount(MountContext context) {
		this.childElement.mount(context);
	}

	@Override
	public void mount(MountContext context) {
		super.mount(context);
	}

	@Override
	public void applyBuiltinSelector(BuiltinSelector selector, Consumer<WidgetInstance> consumer) {
		switch (selector) {
			case NthElement nthElement -> {
				if (nthElement.n() == 0 || nthElement.n() == -1) BuiltinSelector.applyNext(getChildElement(),
						nthElement,
						consumer);
			}
			case Children children -> BuiltinSelector.applyNext(getChildElement(), children, consumer);
		}
	}
}
